/*
**      config.c - Ember config handling routines
**
**
** Copyright (c) 2002  Hughes Technologies Pty Ltd
**
** Permission to use, copy, and distribute for non-commercial purposes,
** is hereby granted without fee, providing that the above copyright
** notice appear in all copies and that both the copyright notice and this
** permission notice appear in supporting documentation.
**
** This software is provided "as is" without any expressed or implied warranty.
**
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "config/config.h"
#include "ember.h"
#include "config.h"
#include "config_priv.h"


/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/

static	int curLine = 0,	/* used for parse error reporting */
	configLoaded = 0;	/* don't double load config files */

static	char errorMessage[1024];


/*
** Configuration table.  Defines what entries are valid, their type,
** and what the default values are
**
**-------------------------------------------------------------------------
** 	entry            type
**	char def         	int def   	  NULL 
*/

static config_entry conf_table [] = {
 {	"inst_dir",		CONFIG_CHAR_TYPE,
	INST_DIR,       	0, 		0
 },
 {	"allow_exec",		CONFIG_INT_TYPE,
	NULL,           	1, 		0
 },
 {	"allow_file_read",	CONFIG_INT_TYPE,
	NULL,           	1, 		0
 },
 {	"allow_file_write",	CONFIG_INT_TYPE,
	NULL,           	1, 		0
 },
 {	"local_files_only",	CONFIG_INT_TYPE,
	NULL,           	0, 		0
 },
 {	"file_path_prefix",	CONFIG_CHAR_TYPE,
	"",           		1, 		0
 },
 {	"exec_path_prefix",	CONFIG_CHAR_TYPE,
	"",           		1, 		0
 },
 { 	NULL,			0,              
	NULL,			0, 		0
 }
};



/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

/*
** Private
** checkIntVal		- Make sure integer values are exactly that
*/
static int checkIntVal(str)
	char	*str;
{
	char 	*cp = str;

	while(*cp)
	{
		if (!isdigit((int)*cp))
			return(-1);
		cp++;
	}
	return(0);
}



/*
** Private
** setConfigEntry	- Update the internal entray value
*/
static int setConfigEntry(handle,value)
	char	*handle,
		*value;
{
	config_entry	*cur;
	char		*cp;

	if (!value)
	{
		sprintf(errorMessage,"Invalid value at line %d", curLine);
		return(CONFIG_ERR_VALUE);
	}
	cur = conf_table;
	while(cur->handle)
	{
		/*
		** Is this the correct entry?
		*/
		if (*cur->handle != *handle || strcmp(cur->handle,handle) != 0)
		{
			cur++;
			continue;
		}

		/*
		** Set the entry value
		*/
		switch(cur->type)
		{
		case CONFIG_CHAR_TYPE:
			if (strcmp(value,"NULL") == 0)
				cur->charVal = NULL;
			else
				cur->charVal = (char *)strdup(value);
			return(CONFIG_OK);
			break;

		case CONFIG_INT_TYPE:
			if (checkIntVal(value) == 0)
			{
				cur->intVal = atoi(value);
			}
			else
			{
				int	goodValue = 0;

				cp = value;
				while(*cp)
				{
					*cp = tolower(*cp);
					cp++;
				}
				if(strcmp(value,"true") == 0)
				{
					cur->intVal = 1;
					goodValue = 1;
				}
				if(strcmp(value,"false") == 0)
				{
					cur->intVal = 0;
					goodValue = 1;
				}
				if (goodValue == 0)
				{
					sprintf(errorMessage,
						"Invalid value at line %d",
						curLine);
					return(CONFIG_ERR_VALUE);
				}
			}
			return(CONFIG_OK);
		}
	}
	snprintf(errorMessage,CONFIG_MAX_ERR_LEN,"Unknown element %s",handle);
	return(CONFIG_ERR_ENTRY);
}


/*
** Private
** processDirective	- process a config file line
*/
static int processDirective(dir, val)
	char	*dir,
		*val;
{
	char	*cp;

	/*
	** Map to lower case
	*/
	cp = dir;
	while(*cp)
	{
		*cp = tolower(*cp);
		cp++;
	}

	/*
	** Do your thing
	*/
	return(setConfigEntry(dir,val));
}



/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/


/*
** Public
** configGetIntEntry	- Return value of integer config entry
*/
int eConfigGetIntEntry(handle)
	char	*handle;
{
	config_entry	*cur;

	cur = conf_table;
	while(cur->handle)
	{
		if (*cur->handle == *handle && strcmp(cur->handle,handle) == 0)
		{
			return(cur->intVal);
		}
		cur++;
	}
	fprintf(stderr,"Invalid w3e config entry '%s'\n", handle);
	return(-1);
}


/*
** Public
** configGetCharEntry	- Return value of character config entry
*/
char * eConfigGetCharEntry(handle)
	char	*handle;
{
	config_entry	*cur;
	static	char nullVal[] = "NULL";

	cur = conf_table;
	while(cur->handle)
	{
		if (*cur->handle==*handle && strcmp(cur->handle,handle) == 0)
		{
			if (cur->charVal == NULL)
			{
				if (cur->allowNull)
					return(NULL);
				else
					return(nullVal);
			}
			return(cur->charVal);
		}
		cur++;
	}
	fprintf(stderr,"Invalid w3e config entry '%s'\n", handle);
	return(NULL);
}


/*
** Public
** configLoadFile	- Load a given or default config file
*/
int eConfigLoadFile(file, section, prefixMatch)
	char	*file,
		*section;
	int	prefixMatch;
{
	FILE	*fp;
	char	buf[256],
		*cp = NULL,
		*directive,
		*value;
	int	inSection = 0,
		defSection = 0,
		errVal = 0,
		fileOK = 1;

	/*
	** Don't double load the config file
	*/
	if (configLoaded == 1 || !file)
	{
		return(CONFIG_ERR_LOADED);
	}

	/*
	** Find and open the config file
	*/
	configLoaded = 1;
	snprintf(buf,sizeof(buf),"%s/%s", INST_DIR, file);
	fp = fopen(buf,"r");
	if (!fp)
	{
		snprintf(buf,sizeof(buf),"%s/%s.conf",INST_DIR,file);
		fp = fopen(buf,"r");
	}
	if (!fp)
	{
		fp = fopen(file,"r");
	}
	if (!fp)
	{
		return(CONFIG_ERR_FILE);
	}


	/*
	** Read and parse the file
	*/
	getLine(buf,160,fp);
	while(!feof(fp))
	{
		/*
		** Dodge blanks
		*/
		cp = buf;
		skipWhite(cp);
		if (*cp == '#' || *cp == '\n' || *cp == '\r')
		{
			getLine(buf,160,fp);
			continue;
		}

		/*
		** Look for a start of section
		*/
		if (*cp == '[')
		{
			if (inSection && ! defSection)
			{
				/* Return after the first section match */
				return(CONFIG_OK);
			}

			directive = (char *)strtok(cp+1," \t]");
			inSection = defSection = 0;
			if (directive[0] == '*' && directive[1] == 0)
			{
				/* Default settings */
				inSection = defSection = 1;
			}
			else
			{
				if (prefixMatch)
				{
					if (strncmp(directive, section,
						strlen(directive)) == 0)
					{
						inSection = 1;
					}
				}
				else
				{
					if (strcmp(directive, section) == 0)
					{
						inSection = 1;
					}
				}
			}
			getLine(buf,160,fp);
			continue;
		}

		if (!inSection)
		{
			getLine(buf,160,fp);
			continue;
		}

		/*
		** Handle directives
		*/
		directive = (char *)strtok(cp, " \t=");
		value = (char *)strtok(NULL, " =\t\n\r");

		errVal =  processDirective(directive,value);
		if (errVal < CONFIG_OK)
		{
			fprintf(stderr,"Config error at line %d : %s\n",
				curLine, errorMessage);
			fileOK = errVal;
		}
		getLine(buf,160,fp);
	}
	fclose( fp);
	if (fileOK != 0)
	{
		return(fileOK);
	}
	else
	{
		return(CONFIG_OK);
	}
}


/*
** Public
** configGetError	- Return a pointer to the last error message
*/
char * eConfigGetError()
{
	return(errorMessage);
}



int eConfigCheckExec(script, command)
	ember	*script;
	char	*command;
{
	char	*cp,
		*prefix;

	if (eConfigGetIntEntry("allow_exec") == 0)
	{
		eRunSetError(script,EMBER_TYPE_INT, 
			"Program execution not allowed");
		return(-1);
	}
	prefix = eConfigGetCharEntry("exec_path_prefix");
	if (prefix && *prefix != 0)
	{
		if (*command != '/')
		{
			eRunSetError(script,EMBER_TYPE_INT,
				"Must specify a complete file path");
			return(-1);
		}
		if (strncmp(prefix, command, strlen(prefix)) != 0)
		{
			eRunSetError(script,EMBER_TYPE_INT,
				"Path does not match prefix.");
			return(-1);
		}
		cp = index(command,'.');
		while(cp)
		{
			if (*(cp+1) == '.')
			{
				eRunSetError(script,EMBER_TYPE_INT,
					"Path cannot contain '..'");
				return(-1);
			}
			cp = index(command,'.');
		}
	}
	return(0);
}


int eConfigCheckFileAccess(script, path, access)
	ember	*script;
	char	*path,
		*access;
{
	char	*cp,
		*prefix;

	if (*access == '<' && eConfigGetIntEntry("allow_file_read") == 0)
	{
		eRunSetError(script,EMBER_TYPE_INT,"File read access denied");
		return(-1);
	}
	if (*access == '>' && eConfigGetIntEntry("allow_file_write") == 0)
	{
		eRunSetError(script,EMBER_TYPE_INT,"File write access denied");
		return(-1);
	}
	if (eConfigGetIntEntry("local_files_only") && index(path,'/'))
	{
		eRunSetError(script,EMBER_TYPE_INT,"Local files only");
		return(-1);
	}
	prefix = eConfigGetCharEntry("file_path_prefix");
	if (prefix && *prefix != 0)
	{ 
		if (*path != '/')
		{
			eRunSetError(script,EMBER_TYPE_INT,
				"Must be a complete file path");
			return(-1);
		}
		if (strncmp(prefix,path,strlen(prefix))!=0)
		{
			eRunSetError(script,EMBER_TYPE_INT,
				"Access denied due to file prefix");
			return(-1);
		}
		cp = index(path,'.');
		while(cp)
		{
			if (*(cp+1) == '.')
			{
				eRunSetError(script,EMBER_TYPE_INT,
					"Path cannot contain '..'");
				return(-1);
			}
		}
	}
	return(0);
}
